open ExtString;;
let (|>) x f = f x;;
let (>>) f g x = g (f x);;
module Bar = Bigarray.Array1;;
open Int64;;
open Dnatree;;
module Dna = DnaTree;;

exception Finish;;

let nat dna = 
  let rec loop n k =
    match Dna.getnext dna with
    | 'P' -> n
    | 'I' | 'F' -> loop n (mul k 2L)
    | 'C' -> loop (add n k) (mul k 2L)
    | _ -> raise Finish
  in loop 0L 1L;;
  
let consts dna =
  let s = Dna.create 100 in
  let rec loop () =
    match Dna.current dna with
    | 'C' -> Dna.append s 'I'; Dna.skip 1 dna; loop ()
    | 'F' -> Dna.append s 'C'; Dna.skip 1 dna; loop ()
    | 'P' -> Dna.append s 'F'; Dna.skip 1 dna; loop ()
    | 'I' when Dna.nth dna 1 = 'C' -> Dna.append s 'P'; Dna.skip 2 dna; loop ()
    | _ -> Dna.commit s         
  in loop (); s;;

type pitem = PBase of char | PSkip of Int64.t | PSearch of Dna.t | POpen | PClose;;

let pattern dna rnafun = 
  let rec loop p lvl =
    match Dna.getnext dna with
    | 'C' -> loop (PBase 'I' :: p) lvl
    | 'F' -> loop (PBase 'C' :: p) lvl
    | 'P' -> loop (PBase 'F' :: p) lvl
    | 'I' -> begin
      match Dna.getnext dna with
      | 'C' -> loop (PBase 'P' :: p) lvl
      | 'P' -> loop (PSkip (nat dna) :: p) lvl
      | 'F' -> Dna.skip 1 dna; loop (PSearch (consts dna) :: p) lvl
      | 'I' -> begin
        match Dna.getnext dna with
        | 'P' -> loop (POpen :: p) (lvl+1)
        | 'C' | 'F' -> if lvl = 0 then p else loop (PClose :: p) (lvl-1)
        | 'I' -> String.init 7 (Dna.nth dna) |> rnafun; Dna.skip 7 dna; loop p lvl
        | _ -> raise Finish     
        end
      | _ -> raise Finish
      end
    | _ -> raise Finish
  in List.rev (loop [] 0);;
  
type titem = TBase of char | TRef of int * int | TLen of int

let template dna rnafun =
  let rec loop t =
    match Dna.getnext dna with
    | 'C' -> loop (TBase 'I' :: t)
    | 'F' -> loop (TBase 'C' :: t)
    | 'P' -> loop (TBase 'F' :: t)
    | 'I' -> begin
      match Dna.getnext dna with
      | 'C' -> loop (TBase 'P' :: t)
      | 'F' | 'P' -> let l = to_int(nat dna) in let n = to_int(nat dna) in loop (TRef(n, l) :: t)
      | 'I' -> begin
        match Dna.getnext dna with
        | 'C' | 'F' -> t
        | 'P' -> let n = to_int(nat dna) in loop (TLen n :: t)
        | 'I' -> String.init 7 (Dna.nth dna) |> rnafun; Dna.skip 7 dna; loop t
        | _ -> raise Finish
        end
      | _ -> raise Finish
      end
    | _ -> raise Finish
  in List.rev (loop []);;

let quote chlist =
  let rec loop dst = function
    | [] -> List.rev dst
    | 'I' :: xs -> loop ('C'::dst) xs
    | 'C' :: xs -> loop ('F'::dst) xs
    | 'F' :: xs -> loop ('P'::dst) xs
    | 'P' :: xs -> loop ('C'::'I'::dst) xs
    | _ -> failwith "quote err"
  in loop [] chlist;;
  
let protect l d =
  if l=0 then d else
  let rec nquote n lst = if n = 0 then lst else nquote (n-1) (quote lst) in
  let qq = Array.make 256 ("", 0) in
  ['I'; 'C'; 'F'; 'P'] |> List.iter (fun c->
    let qs = String.implode (nquote l [c]) in qq.(int_of_char c)<-(qs, String.length qs) );
  let sumlen = ref 0 in
  Dna.iter_int d (fun ic-> let (_,len)=qq.(ic) in sumlen := len + !sumlen);  
  let dna = Dna.create !sumlen in
  Dna.iter_int d (fun ic-> let (qs,_)=qq.(ic) in String.iter (Dna.append dna) qs);
  Dna.commit dna;
  dna;;
    
let asnat n =
  let d = Dna.create 35 in
  let rec loop k = 
    if k = 0L then (Dna.append d 'P'; Dna.commit d; d) else
    (Dna.append d (if rem k 2L = 0L then 'I' else 'C'); loop (div k 2L))
  in loop n;;
  
let domatch pat dna =
  let dnalen = Dna.length dna in
  let rec loop i e c = function
    | [] -> if i<=dnalen then (Dna.skip i dna; Some (List.rev e)) else None
    | PBase ch :: ps -> if i >= dnalen then None else
                        if Dna.nth dna i = ch then loop (i+1) e c ps else None
    | PSkip n :: ps -> if add (of_int i) n > of_int (Dna.length dna) then None else (loop (i+to_int n) e c ps)
    | PSearch d :: ps -> (match Dna.search dna i d with Some j-> loop j e c ps | None->None)
    | POpen :: ps -> loop i e (i::c) ps
    | PClose :: ps -> let chd::ctl = c in loop i (Dna.sub dna chd i :: e) ctl ps
  in loop 0 [] [] pat;;

type ritem = RBase of char | RDna of Dna.t;;

let replace tpl elst dna =
  let e = Array.of_list elst in
  let mpty = Dna.create 1 in
  let r = List.map (function 
    | TBase c -> RBase c
    | TRef(n, l) -> if n >= Array.length e then RDna mpty else RDna (protect l e.(n))
    | TLen n -> let len = if n >= Array.length e then 0 else Dna.length e.(n) in RDna (asnat (of_int len))
    ) tpl in
  let rlen = List.fold_left (fun sum ri-> 
    match ri with RBase _ -> 1+sum | RDna d -> sum + Dna.length d) 0 r in
  let d = Dna.create (Dna.ndnalen (rlen + Dna.length dna)) in
  List.iter (function 
    | RBase c -> Dna.append d c
    | RDna rdna -> Dna.append_dna d rdna) r;
  Dna.append_dna d dna;
  d;;
  
let print_p =
  List.iter (function
    | PBase c -> Printf.printf "%c " c
    | PSkip n -> Printf.printf "!%s " (to_string n)
    | PSearch d -> Printf.printf "?%s " (Dna.to_string d)
    | POpen -> print_string "( "
    | PClose -> print_string ") ");;
  
let print_t = 
  List.iter (function
    | TBase c -> Printf.printf "%c " c
    | TRef(n, l) -> Printf.printf "%d/%d " n l
    | TLen n -> Printf.printf "|%d| " n);; 

let execute srcdna rnafname = 
  let rnafile = open_out rnafname in
  let rnafun = output_string rnafile in
  try
    let rec loop dna ni =
      Dna.compact dna;
      if ni mod 1000 = 0 then Printf.printf "\n%d) cnt=%d " ni (Dna.cnt dna); flush stdout;       
      let p = pattern dna rnafun in
      let t = template dna rnafun in
      match domatch p dna with
      | None -> loop dna (ni+1)
      | Some e-> 
        if ni mod 1000 = 999 then flush rnafile;
        loop (replace t e dna) (ni+1)
    in loop srcdna 1 
  with Finish -> close_out rnafile;  print_string "ok! ";;

Gc.set { (Gc.get()) with Gc.minor_heap_size = 4194304; Gc.major_heap_increment = 1048576;
               Gc.max_overhead = 1000};;

if Array.length Sys.argv < 3 then print_string "usage: endo dna_file rna_file" 
else execute (Dna.of_file Sys.argv.(1)) Sys.argv.(2);;  


